#include "LiquidCrystal_I2C_Ext.h"
#include <inttypes.h>
#include <Arduino.h>
#include <Wire.h>

uint8_t scrollUp[8]  = {0x4, 0xa, 0x11, 0x1f};
uint8_t scrollDown[8]  = {0x0, 0x0, 0x0, 0x0, 0x1f, 0x11, 0xa, 0x4};

uint8_t iconOk[8]  = {0x0, 0x1, 0x1, 0x2, 0x12, 0xc, 0x4};
uint8_t iconCancel[8]  = {0x0, 0x0, 0x11, 0xa, 0x4, 0xa, 0x11};

char numericSigns[] = "- ";

// When the display powers up, it is configured as follows:
//
// 1. Display clear
// 2. Function set:
//    DL = 1; 8-bit interface data
//    N = 0; 1-line display
//    F = 0; 5x8 dot character font
// 3. Display on/off control:
//    D = 0; Display off
//    C = 0; Cursor off
//    B = 0; Blinking off
// 4. Entry mode set:
//    I/D = 1; Increment by 1
//    S = 0; No shift
//
// Note, however, that resetting the Arduino doesn't reset the LCD, so we
// can't assume that its in that state when a sketch starts (and the
// LiquidCrystal constructor is called).

LiquidCrystal_I2C_Ext::LiquidCrystal_I2C_Ext(uint8_t lcd_addr, uint8_t lcd_cols, uint8_t lcd_rows, uint8_t charsize)
{
  _addr = lcd_addr;
  _cols = lcd_cols;
  _rows = lcd_rows;
  _charsize = charsize;
  _backlightval = LCD_BACKLIGHT;
}

void LiquidCrystal_I2C_Ext::begin() {
  Wire.begin();
  _displayfunction = LCD_4BITMODE | LCD_1LINE | LCD_5x8DOTS;

  if (_rows > 1) {
    _displayfunction |= LCD_2LINE;
  }

  // for some 1 line displays you can select a 10 pixel high font
  if ((_charsize != 0) && (_rows == 1)) {
    _displayfunction |= LCD_5x10DOTS;
  }

  // SEE PAGE 45/46 FOR INITIALIZATION SPECIFICATION!
  // according to datasheet, we need at least 40ms after power rises above 2.7V
  // before sending commands. Arduino can turn on way befer 4.5V so we'll wait 50
  delay(50);

  // Now we pull both RS and R/W low to begin commands
  expanderWrite(_backlightval); // reset expanderand turn backlight off (Bit 8 =1)
  delay(1000);

  //put the LCD into 4 bit mode
  // this is according to the hitachi HD44780 datasheet
  // figure 24, pg 46

  // we start in 8bit mode, try to set 4 bit mode
  write4bits(0x03 << 4);
  delayMicroseconds(4500); // wait min 4.1ms

  // second try
  write4bits(0x03 << 4);
  delayMicroseconds(4500); // wait min 4.1ms

  // third go!
  write4bits(0x03 << 4);
  delayMicroseconds(150);

  // finally, set to 4-bit interface
  write4bits(0x02 << 4);

  // set # lines, font size, etc.
  command(LCD_FUNCTIONSET | _displayfunction);

  // turn the display on with no cursor or blinking default
  _displaycontrol = LCD_DISPLAYON | LCD_CURSOROFF | LCD_BLINKOFF;
  display();

  // clear it off
  clear();

  // Initialize to default text direction (for roman languages)
  _displaymode = LCD_ENTRYLEFT | LCD_ENTRYSHIFTDECREMENT;

  // set the entry mode
  command(LCD_ENTRYMODESET | _displaymode);

  home();
}

/********** high level commands, for the user! */
void LiquidCrystal_I2C_Ext::clear() {
  command(LCD_CLEARDISPLAY);// clear display, set cursor position to zero
  delayMicroseconds(2000);  // this command takes a long time!
}

void LiquidCrystal_I2C_Ext::home() {
  command(LCD_RETURNHOME);  // set cursor position to zero
  delayMicroseconds(2000);  // this command takes a long time!
}

void LiquidCrystal_I2C_Ext::setCursor(uint8_t col, uint8_t row) {
  int row_offsets[] = { 0x00, 0x40, 0x14, 0x54 };
  if (row > _rows) {
    row = _rows - 1;  // we count rows starting w/0
  }
  command(LCD_SETDDRAMADDR | (col + row_offsets[row]));
}

// Turn the display on/off (quickly)
void LiquidCrystal_I2C_Ext::noDisplay() {
  _displaycontrol &= ~LCD_DISPLAYON;
  command(LCD_DISPLAYCONTROL | _displaycontrol);
}
void LiquidCrystal_I2C_Ext::display() {
  _displaycontrol |= LCD_DISPLAYON;
  command(LCD_DISPLAYCONTROL | _displaycontrol);
}

// Turns the underline cursor on/off
void LiquidCrystal_I2C_Ext::noCursor() {
  _displaycontrol &= ~LCD_CURSORON;
  command(LCD_DISPLAYCONTROL | _displaycontrol);
}
void LiquidCrystal_I2C_Ext::cursor() {
  _displaycontrol |= LCD_CURSORON;
  command(LCD_DISPLAYCONTROL | _displaycontrol);
}

// Turn on and off the blinking cursor
void LiquidCrystal_I2C_Ext::noBlink() {
  _displaycontrol &= ~LCD_BLINKON;
  command(LCD_DISPLAYCONTROL | _displaycontrol);
}
void LiquidCrystal_I2C_Ext::blink() {
  _displaycontrol |= LCD_BLINKON;
  command(LCD_DISPLAYCONTROL | _displaycontrol);
}

// These commands scroll the display without changing the RAM
void LiquidCrystal_I2C_Ext::scrollDisplayLeft(void) {
  command(LCD_CURSORSHIFT | LCD_DISPLAYMOVE | LCD_MOVELEFT);
}
void LiquidCrystal_I2C_Ext::scrollDisplayRight(void) {
  command(LCD_CURSORSHIFT | LCD_DISPLAYMOVE | LCD_MOVERIGHT);
}

// This is for text that flows Left to Right
void LiquidCrystal_I2C_Ext::leftToRight(void) {
  _displaymode |= LCD_ENTRYLEFT;
  command(LCD_ENTRYMODESET | _displaymode);
}

// This is for text that flows Right to Left
void LiquidCrystal_I2C_Ext::rightToLeft(void) {
  _displaymode &= ~LCD_ENTRYLEFT;
  command(LCD_ENTRYMODESET | _displaymode);
}

// This will 'right justify' text from the cursor
void LiquidCrystal_I2C_Ext::autoscroll(void) {
  _displaymode |= LCD_ENTRYSHIFTINCREMENT;
  command(LCD_ENTRYMODESET | _displaymode);
}

// This will 'left justify' text from the cursor
void LiquidCrystal_I2C_Ext::noAutoscroll(void) {
  _displaymode &= ~LCD_ENTRYSHIFTINCREMENT;
  command(LCD_ENTRYMODESET | _displaymode);
}

// Allows us to fill the first 8 CGRAM locations
// with custom characters
void LiquidCrystal_I2C_Ext::createChar(uint8_t location, uint8_t charmap[]) {
  location &= 0x7; // we only have 8 locations 0-7
  command(LCD_SETCGRAMADDR | (location << 3));
  for (int i = 0; i < 8; i++) {
    write(charmap[i]);
  }
}

// Turn the (optional) backlight off/on
void LiquidCrystal_I2C_Ext::noBacklight(void) {
  _backlightval = LCD_NOBACKLIGHT;
  expanderWrite(0);
}

void LiquidCrystal_I2C_Ext::backlight(void) {
  _backlightval = LCD_BACKLIGHT;
  expanderWrite(0);
}
bool LiquidCrystal_I2C_Ext::getBacklight() {
  return _backlightval == LCD_BACKLIGHT;
}

/*********** extended functions ***********/
// printAt
void LiquidCrystal_I2C_Ext::printAt(uint8_t col, uint8_t row, const String &s) {
  setCursor(col, row);
  print(s);
}

void LiquidCrystal_I2C_Ext::printAt(uint8_t col, uint8_t row, const char str[]) {
  setCursor(col, row);
  print(str);
}

void LiquidCrystal_I2C_Ext::printAt(uint8_t col, uint8_t row, char c) {
  setCursor(col, row);
  print(c);
}

void LiquidCrystal_I2C_Ext::printAt(uint8_t col, uint8_t row, unsigned char b, int base) {
  setCursor(col, row);
  print(b, base);
}

void LiquidCrystal_I2C_Ext::printAt(uint8_t col, uint8_t row, int n, int base) {
  setCursor(col, row);
  print(n, base);
}

void LiquidCrystal_I2C_Ext::printAt(uint8_t col, uint8_t row, unsigned int n, int base) {
  setCursor(col, row);
  print(n, base);
}

void LiquidCrystal_I2C_Ext::printAt(uint8_t col, uint8_t row, long n, int base) {
  setCursor(col, row);
  print(n, base);
}

void LiquidCrystal_I2C_Ext::printAt(uint8_t col, uint8_t row, unsigned long n, int base) {
  setCursor(col, row);
  print(n, base);
}

void LiquidCrystal_I2C_Ext::printAt(uint8_t col, uint8_t row, double d, int digits) {
  setCursor(col, row);
  print(d, digits);
}

void LiquidCrystal_I2C_Ext::printAt(uint8_t col, uint8_t row, const Printable& x) {
  setCursor(col, row);
  print(x);
}

void LiquidCrystal_I2C_Ext::printf(const char * format, ...) {
  char buf[128];
  va_list args;
  va_start (args, format);
  vsnprintf(buf, 128, format, args);
  va_end (args);
  print(buf);
}

void LiquidCrystal_I2C_Ext::printfAt(uint8_t col, uint8_t row, const char * format, ...) {
  char buf[128];
  va_list args;
  va_start (args, format);
  vsnprintf(buf, 128, format, args);
  va_end (args);
  setCursor(col, row);
  print(buf);
}

void LiquidCrystal_I2C_Ext::attachEncoder(uint8_t pinA, uint8_t pinB, uint8_t pinBtn) {
  _pinA = pinA;
  _pinB = pinB;
  _pinBtn = pinBtn;
  pinMode(_pinA, INPUT_PULLUP);
  pinMode(_pinB, INPUT_PULLUP);
  pinMode(_pinBtn, INPUT_PULLUP);
  _prevPoolTime = 0;
}

eEncoderState LiquidCrystal_I2C_Ext::getEncoderState() {
  unsigned long currentTime = millis();
  bool encoderA, encoderB;
  eEncoderState Result = eNone;
  if (currentTime >= (_prevPoolTime + 5)) {
    _prevPoolTime = currentTime;
    if (digitalRead(_pinBtn) == LOW ) {
      if (_pinButtonPrev) {
        _pinButtonPrev = 0;
        Result = eButton;
      }
    }
    else {
      _pinButtonPrev = 1;
      encoderA = digitalRead(_pinA);
      encoderB = digitalRead(_pinB);
      if ((!encoderA) && (_pinAPrev)) {
        if (encoderB) Result = eRight;
        else          Result = eLeft;
      }
      _pinAPrev = encoderA;
    }
  }
  return Result;
}

bool  LiquidCrystal_I2C_Ext::printTitle(const char title[]) {
  uint8_t l = strlen(title);
  char * buffer;
  clear();
  if (l > 0) {
    l = min(l, _cols);
    buffer = (char*) malloc(l + 1);
    memcpy(buffer, title, l);
    buffer[l] = '\0';
    printAt(0, 0, buffer);
    free(buffer);
    return 1;
  }
  else return 0;
}

void LiquidCrystal_I2C_Ext::printMultiline(const String &s) {
  printMultiline(s.c_str());
}

void LiquidCrystal_I2C_Ext::printMultiline(const char str[]) {
  uint8_t offset = 0;
  uint8_t lineLength = _cols - 1;
  bool needRepaint = 1;
  char buf[_cols];
  eEncoderState encoderState = eNone;
  createChar(0, scrollUp);
  createChar(1, scrollDown);
  while (encoderState != eButton) {
    if (needRepaint) {
      needRepaint = 0;
      clear();
      // print text
      for (uint8_t i = 0; i < min(_rows, (strlen(str) + lineLength - 1) / lineLength); i++) {
        memcpy(buf, &str[offset + i * lineLength], lineLength);
        buf[lineLength] = '\0';
        printAt(0, i, buf);
      }
      // show scrollbar
      if (offset > 0) {
        setCursor(lineLength, 0);
        write(0); // arrow up
      }
      if (strlen(str) > offset + _rows * lineLength) {
        setCursor(lineLength, _rows - 1);
        write(1); // arrow down
      }
    }

    encoderState = getEncoderState();
    if (encoderState == eNone) {
      encoderIdle(); // no action
    }
    else if (encoderState == eLeft) {
      // scroll up
      if (offset > 0) {
        offset -= lineLength;
        needRepaint = 1;
      }
    }
    else if (encoderState == eRight) {
      // scroll down
      if (strlen(str) > offset + _rows * lineLength) {
        offset += lineLength;
        needRepaint = 1;
      }
    }
  }
  clear();
}

bool LiquidCrystal_I2C_Ext::isEditable(char ch, char availSymbols[]) {
  if (ch == '\0') return 1;
  byte l = strlen(availSymbols);
  for (byte i = 0; i < l; i++) {
    if (availSymbols[i] == ch) return 1;
  }
  return 0;
}

bool LiquidCrystal_I2C_Ext::getNextEditable(char S[], uint8_t lenS, char availSymbols[], uint8_t &currentPos, bool direction) {
  int i;
  if (currentPos == 254) i = lenS;
  else i = currentPos;
  while (1) {
    if (direction) i++; else i--;
    if (i < 0) return 0;
    if (i == lenS) {
      currentPos = 254;
      return 0;
    }
    if (isEditable(S[i], availSymbols)) {
      currentPos = i;
      return 1;
    }
  }
}

bool LiquidCrystal_I2C_Ext::_inputStrVal(const char title[], char buffer[], uint8_t len, char availSymbols[], bool _signed) {
  eEncoderState encoderState = eNone;
  uint8_t pos = 0;
  uint8_t l = len;
  if (l > _cols - 4) l = _cols - 4;
  char tmpBuffer[_cols + 1];
  uint8_t hasTitle = printTitle(title);

  memcpy(tmpBuffer, buffer, l);
  tmpBuffer[l] = '\0';
  printAt(0, hasTitle, tmpBuffer);

  createChar(0, iconOk);
  createChar(1, iconCancel);
  setCursor(_cols - 3, hasTitle); // Ok
  write(0);
  setCursor(_cols - 1, hasTitle); // Cancel
  write(1);

  //     
  if (!_signed & (!isEditable(tmpBuffer[pos], availSymbols))) {
    if (!getNextEditable(tmpBuffer, len, availSymbols, pos, 1))
      setCursor(_cols - 3, hasTitle);
    else
      setCursor(pos, hasTitle);
  }
  else setCursor(pos, hasTitle);

  cursor();

  //  -      OK/Cancel
  while (1) {
    encoderState = getEncoderState();
    switch (encoderState) {
      case eNone: {
          encoderIdle();
          continue;
        }
      case eLeft: { //   
          if (pos == 0) continue; //   
          if ((pos == 1) & (_signed)) { //    
            setCursor(--pos, hasTitle);
            continue;
          }
          if (pos == 255) { //  Cancel,   OK
            setCursor(_cols - 3, hasTitle);
            pos--;
            continue;
          }
          if (getNextEditable(tmpBuffer, len, availSymbols, pos, 0))
            setCursor(pos, hasTitle);
          continue;
        }
      case eRight: { //   
          if (pos == 255) continue; //   
          if (pos == 254) { //  Ok,   Cancel
            setCursor(_cols - 1, hasTitle);
            pos++;
            continue;
          }
          if (pos == l - 1) { //    ,   OK
            setCursor(_cols - 3, hasTitle);
            pos = 254;
            continue;
          }
          if (getNextEditable(tmpBuffer, len, availSymbols, pos, 1))
            setCursor(pos, hasTitle);
          else {
            pos = 254;
            setCursor(_cols - 3, hasTitle);
          }
          continue;
        }
      case eButton: { // 
          if (pos == 255) {
            noCursor();
            clear();
            return 0; // Cancel.
          }
          if (pos == 254) { // OK.
            noCursor();
            clear();
            memcpy(buffer, tmpBuffer, l);
            return 1;
          }
          //   
          encoderState = eNone;
          //setCursor(pos, hasTitle);
          blink();
          while (encoderState != eButton)
          {
            encoderState = getEncoderState();
            switch (encoderState) {
              case eNone: {
                  encoderIdle();
                  continue;
                }
              case eLeft: {
                  if ((_signed) & (pos == 0)) {
                    if (!getNextSymbol(tmpBuffer[pos], 0, numericSigns, 1)) continue;
                  }
                  else {
                    if (!getNextSymbol(tmpBuffer[pos], 0, availSymbols)) continue;
                  }
                  printAt(pos, hasTitle, tmpBuffer[pos]);
                  setCursor(pos, hasTitle);
                  continue;
                }
              case eRight: {
                  if ((_signed) & (pos == 0)) {
                    if (!getNextSymbol(tmpBuffer[pos], 1, numericSigns, 1)) continue;
                  }
                  else {
                    if (!getNextSymbol(tmpBuffer[pos], 1, availSymbols)) continue;
                  }
                  printAt(pos, hasTitle, tmpBuffer[pos]);
                  setCursor(pos, hasTitle);
                  continue;
                }
            }
          }
          noBlink();
          continue;
        }
    }
  }
}

bool LiquidCrystal_I2C_Ext::getNextSymbol(char &ch, bool direction, char availSymbols[], bool looped) {
  if (ch == '\0') {
    ch = availSymbols[0];
    return 1;
  }
  for (int i = 0; i < strlen(availSymbols); i++) {
    if (availSymbols[i] == ch) {
      if (direction) {
        if (++i < strlen(availSymbols)) ch = availSymbols[i];
        else if (looped) ch = availSymbols[0];
        return 1;
      }
      else {
        if (--i >= 0) ch = availSymbols[i];
        else if (looped) ch = availSymbols[strlen(availSymbols) - 1];
        return 1;
      }
      return 0;
    }
  }
}

bool LiquidCrystal_I2C_Ext::inputStrVal(const String &title, char buffer[], uint8_t len, char availSymbols[]) {
  return _inputStrVal(title.c_str(), buffer, len, availSymbols, 0);
}

bool LiquidCrystal_I2C_Ext::inputStrVal(const char title[], char buffer[], uint8_t len, char availSymbols[]) {
  return _inputStrVal(title, buffer, len, availSymbols, 0);
}

uint8_t LiquidCrystal_I2C_Ext::selectVal(const String &title, char **list, uint8_t count, uint8_t preselected) {
  return _selectVal(title.c_str(), list, count, preselected);
}

uint8_t LiquidCrystal_I2C_Ext::selectVal(const char title[], char **list, uint8_t count, uint8_t preselected) {
  return _selectVal(title, list, count, preselected);
}

uint8_t LiquidCrystal_I2C_Ext::selectVal(const String &title, String list[], uint8_t count, uint8_t preselected) {
  return _selectVal(title.c_str(), list, count, preselected);
}

uint8_t LiquidCrystal_I2C_Ext::selectVal(const char title[], String list[], uint8_t count, uint8_t preselected) {
  return _selectVal(title, list, count, preselected);
}

uint8_t LiquidCrystal_I2C_Ext::selectVal(const String &title, int list[], uint8_t count, uint8_t preselected) {
  return _selectVal(title.c_str(), list, count, preselected);
}

uint8_t LiquidCrystal_I2C_Ext::selectVal(const char title[], int list[], uint8_t count, uint8_t preselected) {
  return _selectVal(title, list, count, preselected);
}

void LiquidCrystal_I2C_Ext::_prepareForPrint(char buffer[], char *value, uint8_t len) {
  char format[12] = {'\0'};
  sprintf(format, "%%-%d.%ds", len, len);
  sprintf(buffer, format, value);
}

void LiquidCrystal_I2C_Ext::_prepareForPrint(char buffer[], int value, uint8_t len) {
  char format[8] = {'\0'};
  sprintf(format, "%%-%dd", len);
  sprintf(buffer, format, value);
}

void LiquidCrystal_I2C_Ext::_prepareForPrint(char buffer[], String value, uint8_t len) {
  char format[12] = {'\0'};
  sprintf(format, "%%-%d.%ds", len, len);
  sprintf(buffer, format, value.c_str());
}

template <typename T>
uint8_t LiquidCrystal_I2C_Ext::_selectVal(const char title[], T list[], uint8_t count, uint8_t preselected) {
  uint8_t offset = 0;
  uint8_t cursorOffset = 0;
  uint8_t lineLength = _cols - 3;
  uint8_t selected = preselected;
  bool needRepaint = 1;
  char buf[_cols + 1];
  uint8_t hasTitle = printTitle(title);

  //     
  createChar(0, scrollUp);
  createChar(1, scrollDown);

  eEncoderState encoderState = eNone;
  while (1) {
    if (needRepaint) {
      needRepaint = 0;
      //   
      for (uint8_t i = 0; i < min(count, _rows - hasTitle); i++) {
        _prepareForPrint(buf, list[offset + i], lineLength);
        printfAt(0, i + hasTitle, "  %s ", buf);
        if (offset + i == selected)
          printAt(1, i + hasTitle, '*');
      }
      //  
      printAt(0, cursorOffset + hasTitle, '>');
      //   
      if (offset > 0) {
        setCursor(_cols - 1, hasTitle);
        write(0); //  
      }
      if (offset + _rows - hasTitle < count) {
        setCursor(_cols - 1, _rows - 1);
        write(1); //  
      }
    }
    encoderState = getEncoderState();
    switch (encoderState) {
      case eNone: { // todo   
          encoderIdle();
          continue;
        }
      case eLeft: {
          if (cursorOffset > 0) {
            printAt(0, cursorOffset + hasTitle, ' ');
            printAt(0, --cursorOffset + hasTitle, '>');
          }
          else if (offset > 0) {
            offset--;
            needRepaint = 1;
          }
          continue;
        }
      case eRight: {
          if (cursorOffset < min(_rows - hasTitle, count) - 1) {
            printAt(0, cursorOffset + hasTitle, ' ');
            printAt(0, ++cursorOffset + hasTitle, '>');
          }
          else if ((cursorOffset + hasTitle + 1 == _rows) & (offset + _rows - hasTitle < count)) {
            offset++;
            needRepaint = 1;
          }
          continue;
        }
      case eButton: {
          if (offset + cursorOffset != selected) {
            // set new selected
            if ((selected >= offset) & (selected < offset + _rows - hasTitle))
              printAt(1, selected - offset + hasTitle, ' ');
            selected = offset + cursorOffset;
            printAt(1, cursorOffset + hasTitle, '*');
            continue;
          }
          else {
            clear();
            return selected;
          }
        }

    }

  }

}

uint8_t LiquidCrystal_I2C_Ext::showMenu(sMenuItem menu[], uint8_t menuLen, bool showTitle) {
  _showMenuTitle = showTitle;
  _menu = menu;
  _menuLen = menuLen;
  uint8_t selectedItem;
  //     
  createChar(0, scrollUp);
  createChar(1, scrollDown);
  selectedItem = showSubMenu(1);
  clear();
  return selectedItem;
}

uint8_t LiquidCrystal_I2C_Ext::showSubMenu(uint8_t key) {
  uint8_t result, subMenuLen = 0, subMenuIndex, offset = 0, cursorOffset = 0;
  sMenuItem **subMenu = NULL;
  bool needRepaint = 1;
  char *buffer;
  eEncoderState encoderState;
  subMenuLen = 0;
  uint8_t itemMaxLength = _cols - 1;
  //   subMenu  
  for (uint8_t i = 0; i < _menuLen; i++) {
    if (_menu[i].key == key)
      subMenuIndex = i;
    else if (_menu[i].parent == key) {
      subMenuLen++;
      subMenu = (sMenuItem**) realloc (subMenu, subMenuLen * sizeof(void*));
      subMenu[subMenuLen - 1] = &_menu[i];
    }
  }

  if (subMenuLen == 0) return key; //  ,     

  //   
  if (subMenuLen > _rows - _showMenuTitle) itemMaxLength--;
  do {
    if (needRepaint) {
      needRepaint = 0;
      clear();
      buffer = (char*) malloc (_cols + 1);
      if (_showMenuTitle) {
        memcpy(buffer, _menu[subMenuIndex].caption, _cols);
        buffer[_cols] = 0;
        printAt(0, 0, buffer);
      }
      buffer[_cols] = 0;
      for (uint8_t i = 0; i < min(subMenuLen, _rows - _showMenuTitle); i++) {
        // " %-*.*s" does not work wtf
        memcpy(buffer, subMenu[offset + i]->caption, itemMaxLength);
        buffer[itemMaxLength] = 0;
        printAt(1 , i + _showMenuTitle,  buffer);
      }

      printAt(0, cursorOffset + _showMenuTitle, '>');

      if (offset > 0) {
        setCursor(_cols - 1, _showMenuTitle);
        write(0);
      }
      if (offset + _rows - _showMenuTitle < subMenuLen) {
        setCursor(_cols - 1, _rows - 1);
        write(1);
      }
      free(buffer);
    }
    encoderState = getEncoderState();
    switch (encoderState) {
      case eLeft: {
          //   
          if (cursorOffset > 0) {  //   ,  
            //     ,   
            printAt(0, cursorOffset + _showMenuTitle, ' ');
            printAt(0, --cursorOffset + _showMenuTitle, '>');
          }
          else if (offset > 0) {
            //    .    ,   
            offset--;
            needRepaint = 1;
          }
          break;
        }
      case eRight: {
          //   
          if (cursorOffset < min(_rows - _showMenuTitle, subMenuLen) - 1) {//   ,   
            //     ,   
            printAt(0, cursorOffset + _showMenuTitle, ' ');
            printAt(0, ++cursorOffset + _showMenuTitle, '>');
          }
          else if ((cursorOffset + _showMenuTitle + 1 == _rows) & (offset + _rows - _showMenuTitle < subMenuLen)) {
            //     .    ,   
            offset++;
            needRepaint = 1;
          }
          break;
        }
      case eButton: {
          //   
          if (subMenu[cursorOffset + offset]->key == 0) { // Back
            free(subMenu);
            return 0;
          }
          //       showSubMenu
          result = showSubMenu(subMenu[cursorOffset + offset]->key);
          if (result != 0) {
            free(subMenu);
            return result;
          }
          needRepaint = 1;
          break;
        }
      case eNone: {// todo   

          encoderIdle();
        }
    }
  } while (1);
}

void LiquidCrystal_I2C_Ext::encoderIdle() {

}


/*********** mid level commands, for sending data/cmds */

inline void LiquidCrystal_I2C_Ext::command(uint8_t value) {
  send(value, 0);
}

inline size_t LiquidCrystal_I2C_Ext::write(uint8_t value) {
  send(value, Rs);
  return 1;
}


/************ low level data pushing commands **********/

// write either command or data
void LiquidCrystal_I2C_Ext::send(uint8_t value, uint8_t mode) {
  uint8_t highnib = value & 0xf0;
  uint8_t lownib = (value << 4) & 0xf0;
  write4bits((highnib) | mode);
  write4bits((lownib) | mode);
}

void LiquidCrystal_I2C_Ext::write4bits(uint8_t value) {
  expanderWrite(value);
  pulseEnable(value);
}

void LiquidCrystal_I2C_Ext::expanderWrite(uint8_t _data) {
  Wire.beginTransmission(_addr);
  Wire.write((int)(_data) | _backlightval);
  Wire.endTransmission();
}

void LiquidCrystal_I2C_Ext::pulseEnable(uint8_t _data) {
  expanderWrite(_data | En);  // En high
  delayMicroseconds(1);   // enable pulse must be >450ns

  expanderWrite(_data & ~En); // En low
  delayMicroseconds(40);    // commands need > 37us to settle
}

void LiquidCrystal_I2C_Ext::load_custom_character(uint8_t char_num, uint8_t *rows) {
  createChar(char_num, rows);
}

void LiquidCrystal_I2C_Ext::setBacklight(uint8_t new_val) {
  if (new_val) {
    backlight();    // turn backlight on
  } else {
    noBacklight();  // turn backlight off
  }
}

void LiquidCrystal_I2C_Ext::printstr(const char c[]) {
  //This function is not identical to the function used for "real" I2C displays
  //it's here so the user sketch doesn't have to be changed
  print(c);
}
